-- Author      : Kurapica
-- Create Date : 2010.10.18
-- Change Log  :

----------------------------------------------------------------------------------------------------------------------------------------
--- ActionButton is inherited from SecureCheckButton with custom style to help author creating useful buttons.
-- <br><br>inherit <a href=".\SecureCheckButton.html">SecureCheckButton</a> For all methods, properties and scriptTypes
-- @name ActionButton
-- @class table
----------------------------------------------------------------------------------------------------------------------------------------

-- Check Version
local version = 1
if not IGAS:NewAddon("IGAS.Widget.ActionButton", version) then
	return
end

class "ActionButton"
	inherit "SecureCheckButton"

	local floor = floor
	local GetBindingKey = GetBindingKey
	local GetBindingText = GetBindingText
	local InCombatLockdown = InCombatLockdown
	local RANGE_INDICATOR = RANGE_INDICATOR

	local SecureButton_GetModifiedAttribute = SecureButton_GetModifiedAttribute
	local SecureButton_GetAttribute = SecureButton_GetAttribute
	local ActionButton_CalculateAction = ActionButton_CalculateAction
	
	-- Register all created ActionButtons
	_ActionButtons = _ActionButtons or setmetatable({}, {__mode = "k"})

	do
		_Manager = _Manager or CreateFrame("SecureFrame", "IGAS_ActionButton_Manager", IGAS.WorldFrame)
		_Manager.Visible = false		
		
		_Manager:RegisterEvent("PLAYER_ENTERING_WORLD")
		_Manager:RegisterEvent("ACTIONBAR_SHOWGRID")
		_Manager:RegisterEvent("ACTIONBAR_HIDEGRID");
		_Manager:RegisterEvent("ACTIONBAR_PAGE_CHANGED")
		_Manager:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
		_Manager:RegisterEvent("UPDATE_BINDINGS")
		_Manager:RegisterEvent("UPDATE_SHAPESHIFT_FORM")
		
		
	end

    -- Scripts
	local _Actions = {
		"actionbar",
		"action",
		"pet",
		"flyout",
		"multispell",
		"spell",
		"glyph",
		"item",
		"macro",
		"cancelaura",
		"stop",
		"target",
		"focus",
		"assist",
		"maintank",
		"mainassist",
		"click",
		"attribute",
	}

	local _ModifiedClks = {
		"SELFCAST",
		"FOCUSCAST",
		"AUTOLOOTTOGGLE",
		"MAILAUTOLOOTTOGGLE",
		"STICKYCAMERA",
		"CHATLINK",
		"DRESSUP",
		"SOCKETITEM",
		"SPLITSTACK",
		"PICKUPACTION",
		"COMPAREITEMS",
		"OPENALLBAGS",
		"QUESTWATCHTOGGLE",
		"TOKENWATCHTOGGLE",
		"SHOWITEMFLYOUT",
		"SHOWMULTICASTFLYOUT",
		"TRADESEARCHADD",
	}

	local _ActionButtonTypes = {
		nil,
		"ACTIONBUTTON",
		"MULTICASTACTIONBUTTON"
	}

	-- Scripts
	local function UpdateHotKey(self)
		local id = self.ID
		local buttonType = self.ButtonType
		local hotkey = self:GetChild("HotKey")

		local key = (buttonType and id and GetBindingKey(buttonType..id)) or
					GetBindingKey("CLICK "..self:GetName()..":LeftButton") or ""

		local text = GetBindingText(key, "KEY_", 1)
		if ( text == "" ) then
			hotkey:SetText(RANGE_INDICATOR)
			hotkey:SetPoint("TOPLEFT", self, "TOPLEFT", 1, -2)
			hotkey:Hide()
		else
			hotkey:SetText(text)
			hotkey:SetPoint("TOPLEFT", self, "TOPLEFT", -2, -2)
			hotkey:Show()
		end
	end

	local function ShowGrid(self)
		self.__ShowGrid = self.__ShowGrid + 1

			if (issecure()) then
				button:SetAttribute("showgrid", button:GetAttribute("showgrid") + 1)
			end

			_G[button:GetName().."NormalTexture"]:SetVertexColor(1.0, 1.0, 1.0, 0.5)

			if ( button:GetAttribute("showgrid") >= 1 and not button:GetAttribute("statehidden") ) then
				button:Show()
			end
	end

	local function HideGrid(self)
		assert(button)

		local showgrid = button:GetAttribute("showgrid")

		if ( issecure() ) then
			if ( showgrid > 0 ) then
				button:SetAttribute("showgrid", showgrid - 1)
			end
		end

		if ( button:GetAttribute("showgrid") == 0 and not HasAction(button.action) ) then
			button:Hide()
		end
	end

	local function UpdateState(self)
		assert(button)

		local action = button.action
		if ( IsCurrentAction(action) or IsAutoRepeatAction(action) ) then
			button:SetChecked(1)
		else
			button:SetChecked(0)
		end
	end

	local function UpdateUsable (self)
		local name = self:GetName()
		local icon = _G[name.."Icon"]
		local normalTexture = _G[name.."NormalTexture"]
		local isUsable, notEnoughMana = IsUsableAction(self.action)
		if ( isUsable ) then
			icon:SetVertexColor(1.0, 1.0, 1.0)
			normalTexture:SetVertexColor(1.0, 1.0, 1.0)
		elseif ( notEnoughMana ) then
			icon:SetVertexColor(0.5, 0.5, 1.0)
			normalTexture:SetVertexColor(0.5, 0.5, 1.0)
		else
			icon:SetVertexColor(0.4, 0.4, 0.4)
			normalTexture:SetVertexColor(1.0, 1.0, 1.0)
		end
	end

	local function UpdateCount (self)
		local text = _G[self:GetName().."Count"]
		local action = self.action
		if ( IsConsumableAction(action) or IsStackableAction(action) ) then
			local count = GetActionCount(action)
			if ( count > (self.maxDisplayCount or 9999 ) ) then
				text:SetText("*")
			else
				text:SetText(count)
			end
		else
			text:SetText("")
		end
	end

	local function UpdateCooldown (self)
		local cooldown = _G[self:GetName().."Cooldown"]
		local start, duration, enable = GetActionCooldown(self.action)
		CooldownFrame_SetTimer(cooldown, start, duration, enable)
	end

	local function Update (self)
		local name = self:GetName()

		local action = self.action
		local icon = _G[name.."Icon"]
		local buttonCooldown = _G[name.."Cooldown"]
		local texture = GetActionTexture(action)

		if ( HasAction(action) ) then
			if ( not self.eventsRegistered ) then
				self:RegisterEvent("ACTIONBAR_UPDATE_STATE")
				self:RegisterEvent("ACTIONBAR_UPDATE_USABLE")
				self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
				self:RegisterEvent("UPDATE_INVENTORY_ALERTS")
				self:RegisterEvent("PLAYER_TARGET_CHANGED")
				self:RegisterEvent("TRADE_SKILL_SHOW")
				self:RegisterEvent("TRADE_SKILL_CLOSE")
				self:RegisterEvent("ARCHAEOLOGY_CLOSED")
				self:RegisterEvent("PLAYER_ENTER_COMBAT")
				self:RegisterEvent("PLAYER_LEAVE_COMBAT")
				self:RegisterEvent("START_AUTOREPEAT_SPELL")
				self:RegisterEvent("STOP_AUTOREPEAT_SPELL")
				self:RegisterEvent("UNIT_ENTERED_VEHICLE")
				self:RegisterEvent("UNIT_EXITED_VEHICLE")
				self:RegisterEvent("COMPANION_UPDATE")
				self:RegisterEvent("UNIT_INVENTORY_CHANGED")
				self:RegisterEvent("LEARNED_SPELL_IN_TAB")
				self:RegisterEvent("PET_STABLE_UPDATE")
				self:RegisterEvent("PET_STABLE_SHOW")
				self:RegisterEvent("SPELL_ACTIVATION_OVERLAY_GLOW_SHOW")
				self:RegisterEvent("SPELL_ACTIVATION_OVERLAY_GLOW_HIDE")
				self.eventsRegistered = true
			end

			if ( not self:GetAttribute("statehidden") ) then
				self:Show()
			end
			UpdateState(self)
			UpdateUsable(self)
			UpdateCooldown(self)
			UpdateFlash(self)
		else
			if ( self.eventsRegistered ) then
				self:UnregisterEvent("ACTIONBAR_UPDATE_STATE")
				self:UnregisterEvent("ACTIONBAR_UPDATE_USABLE")
				self:UnregisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
				self:UnregisterEvent("UPDATE_INVENTORY_ALERTS")
				self:UnregisterEvent("PLAYER_TARGET_CHANGED")
				self:UnregisterEvent("TRADE_SKILL_SHOW")
				self:UnregisterEvent("ARCHAEOLOGY_CLOSED")
				self:UnregisterEvent("TRADE_SKILL_CLOSE")
				self:UnregisterEvent("PLAYER_ENTER_COMBAT")
				self:UnregisterEvent("PLAYER_LEAVE_COMBAT")
				self:UnregisterEvent("START_AUTOREPEAT_SPELL")
				self:UnregisterEvent("STOP_AUTOREPEAT_SPELL")
				self:UnregisterEvent("UNIT_ENTERED_VEHICLE")
				self:UnregisterEvent("UNIT_EXITED_VEHICLE")
				self:UnregisterEvent("COMPANION_UPDATE")
				self:UnregisterEvent("UNIT_INVENTORY_CHANGED")
				self:UnregisterEvent("LEARNED_SPELL_IN_TAB")
				self:UnregisterEvent("PET_STABLE_UPDATE")
				self:UnregisterEvent("PET_STABLE_SHOW")
				self:UnregisterEvent("SPELL_ACTIVATION_OVERLAY_GLOW_SHOW")
				self:UnregisterEvent("SPELL_ACTIVATION_OVERLAY_GLOW_HIDE")
				self.eventsRegistered = nil
			end

			if ( self:GetAttribute("showgrid") == 0 ) then
				self:Hide()
			else
				buttonCooldown:Hide()
			end
		end

		-- Add a green border if button is an equipped item
		local border = _G[name.."Border"]
		if ( IsEquippedAction(action) ) then
			border:SetVertexColor(0, 1.0, 0, 0.35)
			border:Show()
		else
			border:Hide()
		end

		-- Update Action Text
		local actionName = _G[name.."Name"]
		if ( not IsConsumableAction(action) and not IsStackableAction(action) ) then
			actionName:SetText(GetActionText(action))
		else
			actionName:SetText("")
		end

		-- Update icon and hotkey text
		if ( texture ) then
			icon:SetTexture(texture)
			icon:Show()
			self.rangeTimer = -1
			self:SetNormalTexture("Interface\\Buttons\\UI-Quickslot2")
		else
			icon:Hide()
			buttonCooldown:Hide()
			self.rangeTimer = nil
			self:SetNormalTexture("Interface\\Buttons\\UI-Quickslot")
			local hotkey = _G[name.."HotKey"]
			if ( hotkey:GetText() == RANGE_INDICATOR ) then
				hotkey:Hide()
			else
				hotkey:SetVertexColor(0.6, 0.6, 0.6)
			end
		end
		UpdateCount(self)

		-- Update flyout appearance
		UpdateFlyout(self)

		UpdateOverlayGlow(self)

		-- Update tooltip
		if ( GameTooltip:GetOwner() == self ) then
			SetTooltip(self)
		end

		self.feedback_action = action
	end

	local function UpdateAction(self)
		
	end

	------------------------------------------------------
	-- Script
	------------------------------------------------------

	------------------------------------------------------
	-- Method
	------------------------------------------------------

	------------------------------------------------------
	-- Property
	------------------------------------------------------
	-- Action
	property "Action" {
		Get = function(self)
			if self.__ActionButtonType == "action" then
				return ActionButton_CalculateAction(IGAS:GetUI(self))
			end
		end,
		Set = function(self, action)
			assert(not InCombatLockdown(), "The Action value can't be changed during combat.")
			if action then
				self.__ActionButtonType = "action"
				self:SetAttribute("type", "action")
				self:SetAttribute("action", action)
			elseif self.__ActionButtonType == "action" then
				self.__ActionButtonType = nil
				self:SetAttribute("type", nil)
				self:SetAttribute("action", nil)
			end
		end,
		Type = Number + nil,
	}
	-- PetAction
	property "PetAction" {
		Get = function(self)
			if self.__ActionButtonType == "pet" then
				return self:GetAttribute("action")
			end
		end,
		Set = function(self, action)
			assert(not InCombatLockdown(), "The PetAction value can't be changed during combat.")
			if action then
				self.__ActionButtonType = "pet"
				self:SetAttribute("type", "pet")
				self:SetAttribute("action", action)
			elseif self.__ActionButtonType == "pet" then
				self.__ActionButtonType = nil
				self:SetAttribute("type", nil)
				self:SetAttribute("action", nil)
			end
		end,
		Type = Number + nil,
	}
	-- Spell
	property "Spell" {
		Get = function(self)
			if self.__ActionButtonType == "spell" then
				return self:GetAttribute("spell")
			end
		end,
		Set = function(self, spell)
			assert(not InCombatLockdown(), "The Spell value can't be changed during combat.")
			if spell then
				self.__ActionButtonType = "spell"
				self:SetAttribute("type", "spell")
				self:SetAttribute("spell", spell)
			elseif self.__ActionButtonType == "spell" then
				self.__ActionButtonType = nil
				self:SetAttribute("type", nil)
				self:SetAttribute("spell", nil)
			end
		end,
		Type = Number + nil,
	}
	-- Item
	property "Item" {
		Get = function(self)
			if self.__ActionButtonType == "item" then
				return self:GetAttribute("item")
			end
		end,
		Set = function(self, item)
			assert(not InCombatLockdown(), "The Item value can't be changed during combat.")
			if item then
				self.__ActionButtonType = "item"
				self:SetAttribute("type", "item")
				self:SetAttribute("item", item)
			elseif self.__ActionButtonType == "item" then
				self.__ActionButtonType = nil
				self:SetAttribute("type", nil)
				self:SetAttribute("item", nil)
			end
		end,
		Type = String + nil,
	}
	-- Macro
	property "Macro" {
		Get = function(self)
			if self.__ActionButtonType == "macro" then
				return self:GetAttribute("macro")
			end
		end,
		Set = function(self, macro)
			assert(not InCombatLockdown(), "The Macro value can't be changed during combat.")
			if item then
				self.__ActionButtonType = "macro"
				self:SetAttribute("type", "macro")
				self:SetAttribute("macrotext", nil)
				self:SetAttribute("macro", macro)
			elseif self.__ActionButtonType == "macro" then
				self.__ActionButtonType = nil
				self:SetAttribute("type", nil)
				self:SetAttribute("macro", nil)
			end
		end,
		Type = String + Number + nil,
	}
	-- MacroText
	property "MacroText" {
		Get = function(self)
			if self.__ActionButtonType == "macrotext" then
				return self:GetAttribute("macro")
			end
		end,
		Set = function(self, macro)
			assert(not InCombatLockdown(), "The MacroText value can't be changed during combat.")
			if item then
				self.__ActionButtonType = "macrotext"
				self:SetAttribute("type", "macro")
				self:SetAttribute("macro", nil)
				self:SetAttribute("macrotext", macro)
			elseif self.__ActionButtonType == "macrotext" then
				self.__ActionButtonType = nil
				self:SetAttribute("type", nil)
				self:SetAttribute("macrotext", nil)
			end
		end,
		Type = String + nil,
	}
	-- ButtonType
	property "ButtonType" {
		Get = function(self)
			return IGAS:GetUI(self).buttonType
		end,
		Set = function(self, btnType)
			if btnType and strtrim(btnType) == "" then
				btnType = nil
			end

			-- Special settings for original frame
			IGAS:GetUI(self).buttonType = btnType
		end,
		Type = String + nil,
	}
	-- AlwaysShowGrid
	property "AlwaysShowGrid" {
		Get = function(self)
			return self.__AlwaysShowGrid or false
		end,
		Set = function(self, flag)
			self.__AlwaysShowGrid = flag
		end,
		Type = Boolean,
	}
	-- IsBonus
	property "IsBonus" {
		Get = function(self)
			return IGAS:GetUI(self).isBonus or false
		end,
		Set = function(self, flag)
			IGAS:GetUI(self).isBonus = flag
		end,
		Type = Boolean,
	}
	-- AlwaysBonus
	property "AlwaysBonus" {
		Get = function(self)
			return IGAS:GetUI(self).alwaysBonus or false
		end,
		Set = function(self, flag)
			IGAS:GetUI(self).alwaysBonus = flag
		end,
		Type = Boolean,
	}
	-- Locked
	property "Locked" {
		Get = function(self)
			return self:GetAttribute("ActionButton_Locked") or false
		end,
		Set = function(self, flag)
			if not InCombatLockdown() then
				self:SetAttribute("Locked", flag)
			end
		end,
		Type = Boolean,
	}
	-- ReceiveDrag
	property "ReceiveDrag" {
		Get = function(self)
			return not self:GetAttribute("ActionButton_NoReceiveDrag")
		end,
		Set = function(self, flag)
			if not InCombatLockdown() then
				self:SetAttribute("ActionButton_NoReceiveDrag", not flag)
			end
		end,
		Type = Boolean,
	}

	------------------------------------------------------
	-- Constructor
	------------------------------------------------------
    function ActionButton(name, parent)
		local button = CreateFrame(_Inherit, name, parent)

		-- button
		button.Width = 36
		button.Height = 36

		button:SetNormalTexture([[Interface\Buttons\UI-Quickslot2]])
		local t = button:GetNormalTexture()
		t:ClearAllPoints()
		t:SetPoint("TOPLEFT", button, "TOPLEFT", -15, 15)
		t:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 15, -15)

		button:SetPushedTexture([[Interface\Buttons\UI-Quickslot-Depress]])

		button:SetHighlightTexture([[Interface\Buttons\ButtonHilight-Square]], "ADD")

		button:SetCheckedTexture([[Interface\Buttons\CheckButtonHilight]], "ADD")

		--- BACKGROUND
		local icon = CreateFrame("Texture", "Icon", button, "BACKGROUND")
		icon:SetAllPoints(button)

		--- ARTWORK - 1
		local flash = CreateFrame("Texture", "Flash", button, "ARTWORK")
		flash.TexturePath = [[Interface\Buttons\UI-QuickslotRed]]
		flash.Visible = false

		local flyoutBorder = CreateFrame("Texture", "FlyoutBorder", button, "ARTWORK", nil, 1)
		flyoutBorder.TexturePath = [[Interface\Buttons\ActionBarFlyoutButton]]
		flyoutBorder:SetTexCoord(0.01562500, 0.67187500, 0.39843750, 0.72656250)
		flyoutBorder:SetPoint("TOPLEFT", button, "TOPLEFT", -3, 3)
		flyoutBorder:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 3, -3)
		flyoutBorder.Visible = false

		local flyoutBorderShadow = CreateFrame("Texture", "FlyoutBorderShadow", button, "ARTWORK", nil, 1)
		flyoutBorderShadow.TexturePath = [[Interface\Buttons\ActionBarFlyoutButton]]
		flyoutBorderShadow:SetTexCoord(0.01562500, 0.76562500, 0.00781250, 0.38281250)
		flyoutBorderShadow:SetPoint("TOPLEFT", button, "TOPLEFT", -6, 6)
		flyoutBorderShadow:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 6, -6)
		flyoutBorderShadow.Visible = false

		--- ARTWORK - 2
		local flyoutArrow = CreateFrame("Texture", "FlyoutArrow", button, "ARTWORK", nil, 2)
		flyoutArrow.TexturePath = [[Interface\Buttons\ActionBarFlyoutButton]]
		flyoutArrow:SetTexCoord(0.62500000, 0.98437500, 0.74218750, 0.82812500)
		flyoutArrow.Width = 23
		flyoutArrow.Height = 11
		flyoutArrow.Visible = false

		local hotKey = CreateFrame("FontString", "HotKey", button, "ARTWORK", "NumberFontNormalSmallGray")
		hotKey.JustifyH = "RIGHT"
		hotKey.Width = 36
		hotKey.Height = 10
		hotKey:SetPoint("TOPLEFT", button, "TOPLEFT", -2, -2)

		local count = CreateFrame("FontString", "Count", button, "ARTWORK", "NumberFontNormal")
		count.JustifyH = "RIGHT"
		count:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -2, 2)

		-- OVERLAY
		local name = CreateFrame("FontString", "Name", button, "OVERLAY", "GameFontHighlightSmallOutline")
		name.Width = 36
		name.Height = 10
		name:SetPoint("BOTTOM", button, "BOTTOM", 0, 2)

		local border = CreateFrame("Texture", "Border", button, "OVERLAY")
		border.TexturePath = [[Interface\Buttons\UI-ActionButton-Border]]
		border:SetBlendMode("ADD")
		border.Visible = false
		border:SetPoint("TOPLEFT", button, "TOPLEFT", -13, 13)
		border:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 13, -13)

		-- Cooldown
		local cooldown = CreateFrame("Cooldown", "Cooldown",  button)
		cooldown:SetPoint("TOPLEFT", button, "TOPLEFT", 0, -1)
		cooldown:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, -1)
		cooldown.Visible = false

		-- Init
		button.__Flashtime = 0
		button.__ShowGrid = 0
		button:SetAttribute("checkselfcast", true)
		button:SetAttribute("checkfocuscast", true)
		button:SetAttribute("useparent-unit", true)
		button:SetAttribute("useparent-actionpage", true)

		button:RegisterForDrag("LeftButton", "RightButton")
		button:RegisterForClicks("AnyUp")

		-- Snippets
		button:SetAttribute("OnReceiveDrag", [[
			local kind, value, subtype = ...
			if not kind or not value then return false end
			local state = self:GetAttribute("state")
			local buttonType, buttonAction = self:GetAttribute("type"), nil
			-- action buttons can do their magic themself
			-- for all other buttons, we'll need to update the content now
			if buttonType ~= "action" and buttonType ~= "pet" then
				-- We get spell book ids from CursorInfo
				-- Convert them to actual spell ids
				if kind == "spell" then
					-- HACK: Use a lookup table to get the spell id from a spellbook slot
					-- API update won't make it into 4.0 =(
					local sid = G_lookupTable[tonumber(value)]
					if sid then
						value = sid
					else
						-- no valid spell
						print("invalid spell", kind, subtype)
						return false
					end
				elseif kind == "item" and value then
					value = format("item:%d", value)
				end

				-- Get the action that was on the button before
				if buttonType ~= "empty" then
					buttonAction = self:GetAttribute(self:GetAttribute("action_field"))
				end

				-- TODO: validate what kind of action is being fed in here
				-- We can only use a handful of the possible things on the cursor
				-- return false for all those we can't put on buttons

				self:SetAttribute(format("labtype-%d", state), kind)
				self:SetAttribute(format("labaction-%d", state), value)
				-- update internal state
				self:RunAttribute("UpdateState", state)
				-- send a notification to the insecure code
				self:CallMethod("ButtonContentsChanged", state, kind, value)
			else
				-- get the action for (pet-)action buttons
				buttonAction = self:GetAttribute("action")
			end
			return self:RunAttribute("PickupButton", buttonType, buttonAction)
		]])

		_ActionButtons[button] = true
		
		return button
    end
endclass "ActionButton"
