-- Author      		: kurapica.igas@gmail.com
-- Create Date 	: 2011/03/01
-- ChangeLog

local version = 1

if not IGAS:NewAddon("IGAS.Widget", version) then
	return
end

------------------------------------------------------
-- Header for System.Widget.*
------------------------------------------------------
import "System"

namespace "System.Widget"

------------------------------------------------------
-- Enums
------------------------------------------------------
-- FramePoint
enum "FramePoint" {
	"TOPLEFT",
	"TOPRIGHT",
	"BOTTOMLEFT",
	"BOTTOMRIGHT",
	"TOP",
	"BOTTOM",
	"LEFT",
	"RIGHT",
	"CENTER",
}
-- FrameStrata
enum "FrameStrata" {
	"PARENT",
	"BACKGROUND",
	"LOW",
	"MEDIUM",
	"HIGH",
	"DIALOG",
	"FULLSCREEN",
	"FULLSCREEN_DIALOG",
	"TOOLTIP",
}
-- DrawLayer
enum "DrawLayer" {
	"BACKGROUND",
	"BORDER",
	"ARTWORK",
	"OVERLAY",
	"HIGHLIGHT",
}
-- AlphaMode
enum "AlphaMode" {
	"DISABLE",
	"BLEND",
	"ALPHAKEY",
	"ADD",
	"MOD",
}
-- OutLineType
enum "OutLineType" {
	"NONE",
	"NORMAL",
	"THICK",
}
-- JustifyVType
enum "JustifyVType" {
	"TOP",
	"MIDDLE",
	"BOTTOM",
}
-- JustifyHType
enum "JustifyHType" {
	"LEFT",
	"CENTER",
	"RIGHT",
}
-- InsertMode
enum "InsertMode" {
	"TOP",
	"BOTTOM",
}
-- Orientation
enum "Orientation" {
	"HORIZONTAL",
	"VERTICAL",
}
-- AttributeType
enum "AttributeType" {
	"nil",
	"boolean",
	"number",
	"string",
}
-- AnimLoopType
enum "AnimLoopType" {
	"NONE",
	"REPEAT",
	"BOUNCE",
}
-- AnimLoopStateType
enum "AnimLoopStateType" {
	"NONE",
	"FORWARD",
	"REVERSE",
}
-- AnimSmoothType
enum "AnimSmoothType" {
	"NONE",
	"IN",
	"OUT",
	"IN_OUT",
	"OUT_IN",
}
-- AnimCurveType
enum "AnimCurveType" {
	"NONE",
	"SMOOTH",
}
-- AnchorType
enum "AnchorType" {
	"ANCHOR_TOPRIGHT",
	"ANCHOR_RIGHT",
	"ANCHOR_BOTTOMRIGHT",
	"ANCHOR_TOPLEFT",
	"ANCHOR_LEFT",
	"ANCHOR_BOTTOMLEFT",
	"ANCHOR_CURSOR",
	"ANCHOR_PRESERVE",
	"ANCHOR_NONE",
}
-- ButtonState
enum "ButtonStateType" {
	"PUSHED",
	"NORMAL",
}
-- ButtonClickType
enum "ButtonClickType" {
	"LeftButtonUp",
	"RightButtonUp",
	"MiddleButtonUp",
	"Button4Up",
	"Button5Up",
	"LeftButtonDown",
	"RightButtonDown",
	"MiddleButtonDown",
	"Button4Down",
	"Button5Down",
	"AnyUp",
	"AnyDown",
}
-- FontColor
enum "FontColor" {
	NORMAL = "|cffffd200",
	HIGHLIGHT = "|cffffffff",
	RED = "|cffff2020",
	GREEN = "|cff20ff20",
	GRAY = "|cff808080",
	YELLOW = "|cffffff00",
	LIGHTYELLOW = "|cffffff9a",
	ORANGE = "|cffff7f3f",
	ACHIEVEMENT = "|cffffff00",
	CLOSE = "|r",
}

------------------------------------------------------
-- Structs
------------------------------------------------------
-- Dimension
struct "Dimension"
	x = Number
	y = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.x = x:Validate(value.x)
		value.y = y:Validate(value.y)

		return value
	end

	function Dimension(x, y)
		local ok, value = pcall(Validate, {x = x, y = y})

		if ok then
			return value
		else
			error("Usage : Dimension(x, y) - "..value)
		end
	end
endstruct "Dimension"

-- Size
struct "Size"
	width = Number
	height = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.width = width:Validate(value.width)
		value.height = height:Validate(value.height)

		return value
	end

	function Size(width, height)
		local ok, value = pcall(Validate, {width = width, height = height})

		if ok then
			return value
		else
			error("Usage : Size(width, height) - "..value)
		end
	end
endstruct "Size"

-- MinMax
struct "MinMax"
	min = Number
	max = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.min = min:Validate(value.min)
		value.max = max:Validate(value.max)

		assert(value.min <= value.max, "%s.min can't be greater than %s.max.")

		return value
	end

	function MinMax(min, max)
		local ok, value = pcall(Validate, {min = min, max = max})

		if ok then
			return value
		else
			error("Usage : MinMax(min, max) - "..value)
		end
	end
endstruct "MinMax"

-- Inset
struct "Inset"
	left = Number
	right = Number
	top = Number
	bottom = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.left = left:Validate(value.left)
		value.right = right:Validate(value.right)
		value.top = top:Validate(value.top)
		value.bottom = bottom:Validate(value.bottom)

		return value
	end

	function Inset(left, right, top, bottom)
		local ok, value = pcall(Validate, {left = left, right = right, top = top, bottom = bottom})

		if ok then
			return value
		else
			error("Usage : Inset(left, right, top, bottom) - "..value)
		end
	end
endstruct "Inset"

-- ColorFloat
struct "ColorFloat"
	function Validate(value)
		assert(type(value) == "number", ("%s must be a number, got %s."):format("%s", type(value)))
		assert(value >= 0 and value <= 1, ("%s must in [0-1], got %s."):format("%s", tostring(value)))
		return value
	end
endstruct "ColorFloat"

-- ColorType
struct "ColorType"
	r = ColorFloat
	g = ColorFloat
	b = ColorFloat
	a = ColorFloat + nil

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.r = r:Validate(value.r)
		value.g = g:Validate(value.g)
		value.b = b:Validate(value.b)
		value.a = a:Validate(value.a) or 1	-- default

		return value
	end

	function ColorType(r, g, b, a)
		local ok, value = pcall(Validate, {r = r, g = g, b = b, a = a})

		if ok then
			return value
		else
			error("Usage : ColorType(r, g, b[, a]) - "..value)
		end
	end
endstruct "ColorType"

-- Position
struct "Position"
	x = Number
	y = Number
	z = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.x = x:Validate(value.x)
		value.y = y:Validate(value.y)
		value.z = y:Validate(value.z)

		return value
	end

	function Position(x, y, z)
		local ok, value = pcall(Validate, {x = x, y = y, z = z})

		if ok then
			return value
		else
			error("Usage : Position(x, y, z) - "..value)
		end
	end
endstruct "Position"

-- FontType
struct "FontType"
	path = String
	height = Number
	outline = OutLineType
	monochrome = Boolean

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.path = path:Validate(value.path)
		value.height = height:Validate(value.height)
		value.outline = outline:Validate(value.outline)
		value.monochrome = monochrome:Validate(value.monochrome)

		return value
	end

	function FontType(path, height, outline, monochrome)
		local ok, value = pcall(Validate, {path = path, height = height, outline = outline, monochrome = monochrome})

		if ok then
			return value
		else
			error("Usage : FontType(path, height, outline, monochrome) - "..value)
		end
	end
endstruct "FontType"

-- BackdropType
struct "BackdropType"
	bgFile = String + nil
	edgeFile = String + nil
	tile = Boolean + nil
	tileSize = Number + nil
	edgeSize = Number + nil
	insets = Inset + nil

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.bgFile = bgFile:Validate(value.bgFile)
		value.edgeFile = edgeFile:Validate(value.edgeFile)
		value.tile = tile:Validate(value.tile)
		value.tileSize = tileSize:Validate(value.tileSize)
		value.edgeSize = edgeSize:Validate(value.edgeSize)
		value.insets = insets:Validate(value.insets)

		return value
	end

	function BackdropType(bgFile, edgeFile, tile, tileSize, edgeSize, insets)
		local ok, value = pcall(Validate, {bgFile = bgFile, edgeFile = edgeFile, tile = tile, tileSize = tileSize, edgeSize = edgeSize, insets = insets})

		if ok then
			return value
		else
			error("Usage : BackdropType(bgFile, edgeFile, tile, tileSize, edgeSize, insets) - "..value)
		end
	end
endstruct "BackdropType"

-- AnimOrderType
struct "AnimOrderType"
	local floor = math.floor

	function Validate(value)
		assert(type(value) == "number", ("%s must be a number, got %s."):format("%s", type(value)))
		assert(value >=0 and value <= 100, ("%s must be in [0-100], got %s."):format("%s", tostring(value)))

		return floor(value)
	end
endstruct "AnimOrderType"

-- AnimOriginType
struct "AnimOriginType"
	point = FramePoint
	x = Number
	y = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.point = point:Validate(value.point)
		value.x = x:Validate(value.x)
		value.y = y:Validate(value.y)

		return value
	end

	function AnimOriginType(point, x, y)
		local ok, value = pcall(Validate, {point = point, x = x, y = y})

		if ok then
			return value
		else
			error("Usage : AnimOriginType(point, x, y) - "..value)
		end
	end
endstruct "AnimOriginType"

-- LightType
struct "LightType"
	enabled = Boolean
	omni = Number
	dirX = Number
	dirY = Number
	dirZ = Number
	ambIntensity = ColorFloat + nil
	ambR = ColorFloat + nil
	ambG = ColorFloat + nil
	ambB = ColorFloat + nil
	dirIntensity = ColorFloat + nil
	dirR = ColorFloat + nil
	dirG = ColorFloat + nil
	dirB = ColorFloat + nil

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.enabled = enabled:Validate(value.enabled)
		value.omni = omni:Validate(value.omni)
		assert(value.omni == 0 or value.omni == 1, "%s.omni must be 0 or 1.")
		value.dirX = dirX:Validate(value.dirX)
		value.dirY = dirY:Validate(value.dirY)
		value.dirZ = dirZ:Validate(value.dirZ)
		value.ambIntensity = ambIntensity:Validate(value.ambIntensity)
		value.ambR = ambR:Validate(value.ambR)
		value.ambG = ambG:Validate(value.ambG)
		value.ambB = ambB:Validate(value.ambB)
		value.dirIntensity = dirIntensity:Validate(value.dirIntensity)
		value.dirR = dirR:Validate(value.dirR)
		value.dirG = dirG:Validate(value.dirG)
		value.dirB = dirB:Validate(value.dirB)

		return value
	end

	function LightType(enabled, omni, dirX, dirY, dirZ, ambIntensity, ambR, ambG, ambB, dirIntensity, dirR, dirG, dirB)
		local ok, value = pcall(Validate, {enabled = enabled, omni =  omni, dirX =  dirX, dirY =  dirY, dirZ =  dirZ, ambIntensity =  ambIntensity, ambR =  ambR, ambG =  ambG, ambB =  ambB, dirIntensity =  dirIntensity, dirR =  dirR, dirG =  dirG, dirB = dirB})

		if ok then
			return value
		else
			error("Usage : LightType(enabled, omni, dirX, dirY, dirZ[, ambIntensity, ambR, ambG, ambB[, dirIntensity, dirR, dirG, dirB]]) - "..value)
		end
	end
endstruct "LightType"

-- MiniMapPosition
struct "MiniMapPosition"
	radius = Number
	rounding = Number
	angel = Number

	function Validate(value)
		assert(type(value) == "table", ("%s must be a table, got %s."):format("%s", type(value)))

		value.radius = radius:Validate(value.radius)
		assert(value.radius > 0, "The %s.radius must be greater than 0.")
		value.rounding = rounding:Validate(value.rounding)
		value.angel = angel:Validate(value.angel)
		assert(value.angel >= 0 and value.angel <= 360, "The %s.angel must in [0-360].")

		return value
	end

	function MiniMapPosition(radius, rounding, angel)
		local ok, value = pcall(Validate, {radius = radius, rounding = rounding, angel = angel})

		if ok then
			return value
		else
			error("Usage : MiniMapPosition(radius, rounding, angel) - "..value)
		end
	end
endstruct "MiniMapPosition"

------------------------------------------------------
-- Global Settings
------------------------------------------------------
------------------------------------
--- Get the true frame of a IGAS frame
-- @name IGAS:GetUI
-- @class function
-- @param frame the IGAS frame
-- @return the true frame of the IGAS frame
-- @usage IGAS:GetUI(MyFrame1)
------------------------------------
function IGAS:GetUI(frame)
	if frame == nil or type(frame) ~= "table" then
		return frame
	else
		return frame.__UI or frame
	end
end

------------------------------------
--- Get the IGAS frame of a frame
-- @name IGAS:GetWrapper
-- @class function
-- @param frame the frame
-- @return the IGAS frame of the frame
-- @usage IGAS:GetWrapper(UIParent)
------------------------------------
function IGAS:GetWrapper(frame)
	if type(frame) ~= "table" or type(frame[0]) ~= "userdata" then
		-- VirtualUIObject's instance will be return here.
		return frame
	end

	if Object.IsClass(frame, Widget["UIObject"]) then
		-- UIObject's instance will be return here.
		return frame
	end

	if frame.__Wrapper and Object.IsClass(frame.__Wrapper, Widget["UIObject"]) then
		-- Check if the frame already has a wrapper
		-- Don't store it because I don't want a hashtable to be refreshed frequently.
		return frame.__Wrapper
	end

	-- Now build a new wrapper for it.
	if not frame.GetObjectType then
		return frame
	end

	local objType = frame:GetObjectType() or "UIObject"

	if Widget[objType] then
		return Widget[objType](frame) or frame
	else
		return Widget["UIObject"](frame) or frame
	end
end

_BaseFrame = _BaseFrame or CreateFrame("Frame")
_BaseFrame:Hide()

------------------------------------
--- Store blz UI's methodes to IGAS's Widget
-- @name StoreBlzMethod
-- @class function
-- @param super the super class
-- @param sample the blz UI element's instance
-- @usage StoreBlzMethod(Region, CreateFrame"Frame"))
------------------------------------
function StoreBlzMethod(cls, managerCls)
	local clsEnv = getfenv(2)
	local name, _, _, super = IGAS:GetClassInfo(cls)

	local sample, manageSample

	if managerCls and IGAS:GetClassInfo(managerCls) then
		local mname = IGAS:GetClassInfo(managerCls)
		manageSample = managerCls(mname, _BaseFrame)
		sample = cls(name, manageSample)
	elseif managerCls then
		sample = managerCls
	else
		sample = cls(name, _BaseFrame)
	end

	if type(IGAS:GetUI(sample)) ~= "table" or getmetatable(IGAS:GetUI(sample)) == nil or type(getmetatable(IGAS:GetUI(sample)).__index) ~= "table" then
		return
	end

	for fname, func in pairs(getmetatable(IGAS:GetUI(sample)).__index) do
		if rawget(clsEnv, fname) == nil and super[fname] == nil then
			clsEnv[fname] = func
		end
	end

	if sample ~= managerCls then
		sample:Dispose()
	end

	if manageSample then
		manageSample:Dispose()
	end
end